"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.sendTransactionBatch = sendTransactionBatch;
exports.sendTransaction = sendTransaction;
const viem_1 = require("viem");
const AccountFactory_js_1 = require("../abis/AccountFactory.js");
const constants_js_1 = require("../constants.js");
const utils_js_1 = require("../utils.js");
const sendPrivyTransaction_js_1 = require("./sendPrivyTransaction.js");
const sendTransactionInternal_js_1 = require("./sendTransactionInternal.js");
async function sendTransactionBatch(client, signerClient, publicClient, parameters, isPrivyCrossApp = false, customPaymasterHandler = undefined) {
    const { calls, paymaster, paymasterInput, ...rest } = parameters;
    if (calls.length === 0) {
        throw new Error('No calls provided');
    }
    if (isPrivyCrossApp) {
        return await (0, sendPrivyTransaction_js_1.sendPrivyTransaction)(client, parameters);
    }
    const batchCalls = parameters.calls.map((tx) => {
        if (!tx.to)
            throw new Error('Transaction target (to) is required');
        return {
            target: tx.to,
            allowFailure: false,
            value: BigInt(tx.value ?? 0),
            callData: tx.data ?? '0x',
        };
    });
    const batchCallData = (0, viem_1.encodeFunctionData)({
        abi: [
            {
                name: 'batchCall',
                type: 'function',
                inputs: [
                    {
                        type: 'tuple[]',
                        name: 'calls',
                        components: [
                            { name: 'target', type: 'address' },
                            { name: 'allowFailure', type: 'bool' },
                            { name: 'value', type: 'uint256' },
                            { name: 'callData', type: 'bytes' },
                        ],
                    },
                ],
                outputs: [],
            },
        ],
        args: [batchCalls],
    });
    const totalValue = batchCalls.reduce((sum, call) => sum + BigInt(call.value), BigInt(0));
    let batchTransaction;
    const isDeployed = await (0, utils_js_1.isSmartAccountDeployed)(publicClient, client.account.address);
    if (!isDeployed) {
        const initialCall = {
            target: client.account.address,
            allowFailure: false,
            value: totalValue,
            callData: batchCallData,
        };
        const initializerCallData = (0, utils_js_1.getInitializerCalldata)(signerClient.account.address, constants_js_1.EOA_VALIDATOR_ADDRESS, initialCall);
        const addressBytes = (0, viem_1.toBytes)(signerClient.account.address);
        const salt = (0, viem_1.keccak256)(addressBytes);
        const deploymentCalldata = (0, viem_1.encodeFunctionData)({
            abi: AccountFactory_js_1.default,
            functionName: 'deployAccount',
            args: [salt, initializerCallData],
        });
        batchTransaction = {
            to: constants_js_1.SMART_ACCOUNT_FACTORY_ADDRESS,
            data: deploymentCalldata,
            value: totalValue,
            paymaster: paymaster,
            paymasterInput: paymasterInput,
            type: 'eip712',
        };
    }
    else {
        batchTransaction = {
            to: client.account.address,
            data: batchCallData,
            value: totalValue,
            paymaster: paymaster,
            paymasterInput: paymasterInput,
            type: 'eip712',
        };
    }
    return (0, sendTransactionInternal_js_1.sendTransactionInternal)(client, signerClient, publicClient, {
        ...batchTransaction,
        ...rest,
    }, constants_js_1.EOA_VALIDATOR_ADDRESS, !isDeployed, {}, customPaymasterHandler);
}
async function sendTransaction(client, signerClient, publicClient, parameters, isPrivyCrossApp = false, customPaymasterHandler = undefined) {
    if (isPrivyCrossApp) {
        let paymasterData = {};
        const requestAsAny = parameters;
        if (customPaymasterHandler &&
            !requestAsAny.paymaster &&
            !requestAsAny.paymasterInput) {
            paymasterData = await customPaymasterHandler({
                ...parameters,
                from: client.account.address,
                chainId: parameters.chain?.id ?? client.chain.id,
            });
        }
        const updatedParameters = {
            ...parameters,
            ...paymasterData,
        };
        return await (0, sendPrivyTransaction_js_1.sendPrivyTransaction)(client, updatedParameters);
    }
    const isDeployed = await (0, utils_js_1.isSmartAccountDeployed)(publicClient, client.account.address);
    if (!isDeployed) {
        const initialCall = {
            target: parameters.to,
            allowFailure: false,
            value: parameters.value ?? 0,
            callData: parameters.data ?? '0x',
        };
        const initializerCallData = (0, utils_js_1.getInitializerCalldata)(signerClient.account.address, constants_js_1.EOA_VALIDATOR_ADDRESS, initialCall);
        const addressBytes = (0, viem_1.toBytes)(signerClient.account.address);
        const salt = (0, viem_1.keccak256)(addressBytes);
        const deploymentCalldata = (0, viem_1.encodeFunctionData)({
            abi: AccountFactory_js_1.default,
            functionName: 'deployAccount',
            args: [salt, initializerCallData],
        });
        parameters.to = constants_js_1.SMART_ACCOUNT_FACTORY_ADDRESS;
        parameters.data = deploymentCalldata;
    }
    return (0, sendTransactionInternal_js_1.sendTransactionInternal)(client, signerClient, publicClient, parameters, constants_js_1.EOA_VALIDATOR_ADDRESS, !isDeployed, {}, customPaymasterHandler);
}
//# sourceMappingURL=sendTransaction.js.map